(*********************************************************************)
(*                          Litmus                                   *)
(*                                                                   *)
(*        Luc Maranget, INRIA Paris-Rocquencourt, France.            *)
(*        Susmit Sarkar, University of Cambridge, UK.                *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique and the authors. All rights reserved.             *)
(*  This file is distributed  under the terms of the Lesser GNU      *)
(*  General Public License.                                          *)
(*********************************************************************)

open Globals
open Printf
let pgm = if Array.length Sys.argv > 0 then Sys.argv.(0) else "litmus"

let parse_int opt s =  match Misc.string_of_intkm s with
| Some x -> x
| None ->    
    raise
      (Arg.Bad
         (sprintf
            "wrong argument '%s'; option '%s' expects an integer argument, possibly suffixed by k or m" s opt))
        
let argint opt f usage =
  opt,Arg.String (fun s -> f (parse_int opt s)),
  sprintf "<n[kKmM]?> %s" usage

let opts =
  [
(* General behavior *)
   "-v", Arg.Unit (fun () -> incr Misc.verbose), " be verbose";
   "-version", Arg.Unit (fun () -> print_endline Config.version; exit 0),
   " show version number and exit";
   "-libdir", Arg.Unit (fun () -> print_endline Config.libdir; exit 0),
   " show installation directory and exit";
   "-switch", Arg.Set Misc.switch, "switch something" ;
   "-o", Arg.String set_tar,
     "<name> cross compilation to directory or tar file <name>" ;
   "-cross",  Arg.String set_cross, "<name> same as -o above, with README and Makefile";
   "-crossrun", Arg.String set_crossrun, "<host> run tests on distant machine <host>, default disabled";
   "-mach", Arg.String MyName.read_cfg,
   "<name> read configuration file name.cfg";
   "-runname", Arg.String set_runname,
   "<name> set run name (for condition selection)";
   "-index",  Arg.String set_exp,
   "<@name> save index of compiled tests in file <@name>" ;
(* Test parameters *)
   "-a", Arg.Int set_avail,
     "<n> Run maximal number of tests concurrently for n available cores (default, run one test)";
   "-avail", Arg.Int set_avail, "<n> alias for -a <n>";

   "-limit", Arg.Bool set_limit,
   sprintf
     "<bool> only run tests with n threads or less (default %b)"
     (get_limit ());

   "-no", Arg.String set_no,
   sprintf
     "<name> do not handle tests whose edge cycles are listed in file <name>";

   "-hint", Arg.String set_hint,
   sprintf
     "<name> read hints in <name>";

   argint "-s" set_size
   (sprintf
      "size of test (default %i)"
      (default_size ())) ;
   "-size_of_test",  Arg.Int set_size,"<n> alias for -s <n>";

   argint "-r" set_run
     (sprintf
        "number of runs (default %i)"
        (default_run ())) ;
   argint "-number_of_run" set_run "alias for -r <n>" ;

(* Modes *)
   "-barrier",
   Arg.String set_barrier,
   sprintf
     "(user|pthread|none|timebase) set type of barriers (default %s)"
     (default_barrier ()) ;

   argint "-delay" set_delay
   (sprintf
     "set timebase delay (default X86=%i, PPC=%i)"
     (default_delay Misc.X86)
     (default_delay Misc.PPC)) ;

   "-vb",
   Arg.Bool set_verbose_barrier,         
     "<b> show time of loop synchronisation (default false)" ;

   "-launch",
   Arg.String set_launch,
   sprintf
     "(changing|fixed) set type of phread lauch (default %s)"
     (default_launch ()) ;
   
   "-mem",
   Arg.String set_memory,
   sprintf
     "(direct|indirect) set type of memory access (default %s)"
     (default_memory ()) ;

   "-contiguous",
   Arg.Bool set_contiguous,
   "<bool> allocate shared locations as a big chunk of memory (default false)" ;

   "-st",
   Arg.Int set_stride,
   sprintf "<n> stride for scanning memory (default %s)"
     (pp_stride (default_stride ())) ;

   "-stride", Arg.Int set_stride, "<n> alias for -st <n>" ;

   "-preload", Arg.String set_preload,
   sprintf "<no|random|custom> preload mode (default %s)" (default_preload ()) ;

   "-collect", Arg.String set_collect,
   sprintf "<local|after|both> collect mode (default %s)" (default_collect ()) ;

   "-safer",
   Arg.String set_safer,
   sprintf "<no|all|write> safer mode (default %s)" (default_safer ()) ;
     
   "-para",
   Arg.String set_para,
   sprintf
     "(self|shell) in cross compilation scripts, technique to run several instances: as posix threads or unix processors (default %s)"
     (default_para ()) ;

   "-affinity",
   Arg.String set_affinity,
   sprintf "(none|incr<n>) attach threads to logical processors (default %s)"
     (default_affinity ()) ;
   "-i", Arg.Int set_affinity_incr,  "<n> alias for -affinity incr<n>" ;

   "-procs",
   Arg.String (fun s -> set_logical_procs s),
   "<int list> specify order of logical processors, implies -affinity incr1" ;

   "-p",Arg.String (fun s -> set_logical_procs s),
   "<int list> alias for -procs <int list>" ;

   "-force_affinity",
   Arg.Bool set_force_affinity,
   sprintf "<bool> force affinity setting, useful to circumvent power management (default %b)"
     default_force_affinity ;

   "-smt_mode",
   Arg.String set_smt_mode,
   sprintf "<none|seq|end> mode for numbering logical processors on the same core, default %s" (default_smt_mode ()) ;
   "-smt",
   Arg.Int set_smt,
   sprintf "<int> specify <int>-ways SMT, default %i" (default_smt ()) ;

   "-cache", Arg.Bool set_cache,
   sprintf "<bool> enable/disable thread cache (default %b)"
     (default_cache ()) ;

   "-call", Arg.Bool set_call,
   sprintf "<bool> enable/disable calling one thread in place of forking it (default %b)"
     (default_call ()) ;
     
   "-prealloc", Arg.Bool set_prealloc,
   sprintf "<bool> alloc memory before forking any thread  (default %b)"
     (default_prealloc ()) ;
     
   "-speedcheck",Arg.String set_speedcheck,
   sprintf
     "<no|some|all> stop test as soon as condition is settled (default %s)"
     (pp_speedcheck (default_speedcheck ())) ;

   "-ccopts", Arg.String set_gcc_opts,
   sprintf
     "<flags> set gcc compilation flags (default X86=\"%s\", PPC=\"%s\", ARM=\"%s\")"
     (default_gcc_opts Misc.X86)
     (default_gcc_opts Misc.ARM)
     (default_gcc_opts Misc.PPC) ;

   "-gcc", Arg.String set_gcc,
   sprintf "<name> name of gcc, default %s"
     (default_gcc ()) ;

   "-linkopt", Arg.String set_link_opt,"<flags> set gcc link option(s)" ;

   "-gas", Arg.Bool set_gas, "<bool> emit Gnu as extensions (default Linux/Mac=true, AIX=false)" ;

(* Specify target machine *)
   "-os",
   Arg.String set_os,
   sprintf
     "(linux|mac|aix) target operating system (default %s)"
     (default_os ()) ;

   "-ws", Arg.String set_word,
   sprintf
     "(w32|w64) word size (default %s)"
     (default_ws ()) ;

(* Misc *)

(*******************)
(* Various timings *)
(*******************)

   "-tsc", Arg.Bool set_tsc, "<bool> get timing info from timestamp counter" ;
   "-clocktime", Arg.Bool set_clocktime, "<bool> get timing info from POSIX clock_gettime function" ;
   "-prof", Arg.Bool set_profiling, "<bool> profile with timestamp counter" ;
   argint "-loop" set_time_loop
     "insert assembly code in a loop of size <n> (default disabled)" ;
   argint "-l" set_time_loop "alias for -l <n>" ;

(* Bizarre *)
   "-isync", Arg.Bool set_isync,
     sprintf "<b> undocumented (default %b)" (default_isync ()); 
   "-syncmacro", Arg.Int set_sync_macro, "<n> undocumented" ;
   "-signaling", Arg.Bool set_signaling_write, "<bool> undocumented" ;
   "-kind", Arg.Bool set_kind, "<bool> undocumented" ;
   "-sleep", Arg.Int set_sleep, "<n> sleep n seconds between tests" ;
   "-xy", Arg.Bool set_xy,"<bool> undocumented";
  ]


let usage = sprintf   "Usage: %s [opts]* filename" pgm

let sources = ref []

let () = Arg.parse opts (fun s -> sources := s :: !sources) usage 

let sources = !sources

let () =
  try
    let outname =
      if Globals.is_out () then Globals.get_tar ()
      else MySys.mktmpdir () in
    let module Tar =
      Tar.Make
        (struct
          let verbose = !Misc.verbose
          let outname = Some outname
        end) in
    let module Top = Top.Top(Tar) in
    Top.from_files sources ;
    if not (Globals.is_out ()) then MySys.rmdir outname ;
    exit 0
  with Misc.Fatal msg ->
    eprintf "Fatal error: %s\n%!" msg ;
    exit 2

