(*********************************************************************)
(*                          Litmus                                   *)
(*                                                                   *)
(*        Luc Maranget, INRIA Paris-Rocquencourt, France.            *)
(*        Susmit Sarkar, University of Cambridge, UK.                *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique and the authors. All rights reserved.             *)
(*  This file is distributed  under the terms of the Lesser GNU      *)
(*  General Public License.                                          *)
(*********************************************************************)

(***********************************************)
(* Parse a source file for the needs of litmus *)
(* (And then compile test)                     *)
(***********************************************)

open Printf

type answer =
  | Completed of Misc.arch * string * string list * StringSet.t
  | Interrupted of Misc.arch * exn
  | Absent of Misc.arch

module Make (A:Arch.S) (Tar:Tar.S)
    (L:GenParser.LexParse with type instruction = A.pseudo)
    (XXXComp : XXXCompile.S with module A = A) =
  struct
    module P = GenParser.Make(A) (L)
    module T = Test.Make(A)
    module Comp = Compile.Make (T) (XXXComp)
    module S = Skel.Make(T)
    module R = Run.Make(Tar)(T)

    let get_cycle t =
      let info = t.MiscParser.info in
      List.assoc "Cycle" info

    let cycle_ok avoid t =
      try
        let cy = get_cycle t in
        not (avoid cy)
      with Not_found -> true

    let is_cycle_new cycles t =
      try
        let cy = get_cycle t in
        if StringSet.mem cy cycles then
          false,cycles
        else
          true,StringSet.add cy cycles
      with Not_found -> true,cycles

    let change_hint hint name t =
      try
        let more_info = Hint.get hint name in
        let info = 
          more_info @
          List.filter
            (fun (k,_) ->
              try
                let _ = List.assoc k more_info in
                false
              with Not_found -> true)
            t.MiscParser.info in
        { t with MiscParser.info = info; }
      with Not_found -> t

    let compile hint avoid_cycle utils cycles name in_chan out_chan splitted  =
      try begin
        let parsed = P.parse in_chan splitted
        and doc = splitted.Splitter.name in
        close_in in_chan ;
        let nprocs = List.length parsed.MiscParser.prog
        and avail = match Globals.get_avail A.arch with
        | Some n -> n
        | None -> 1000 in
        if
          cycle_ok avoid_cycle parsed &&
          (not (Globals.get_limit ()) || nprocs <= avail)
        then begin
          let parsed = change_hint hint doc.Name.name parsed in
          let module Alloc = SymbReg.Make(A) in
          let allocated = Alloc.allocate_regs parsed in
          let compiled = Comp.compile allocated in
          let source = MyName.outname name ".c" in
          let () =
            Misc.output_protect
              (S.dump doc compiled)
              (Tar.outname source) in
          let utils =
            match utils with
            | [] ->
                let module O = ObjUtil.Make(A)(Tar) in
                O.dump ()
            | _ -> utils  in
          R.run name out_chan doc allocated source ;
          Completed (A.arch,source,utils,cycles)
        end else begin
          Warn.warn
            "Test %s not performed" doc.Name.name ;
          Absent A.arch
        end
      end with e -> Interrupted (A.arch,e)

  end

module Top(Tar:Tar.S) =
  struct
    let from_chan hint avoid_cycle fst cycles name in_chan out_chan =  
(* First split the input file in sections *)
      let { Splitter.arch=arch ; _ } as splitted =  Splitter.split name in_chan in
(* Then call appropriate compiler, depending upon arch *)
      let module V = SymbConstant in
      match arch with
      | Misc.PPC ->
          let module Conf = struct
            let word = Globals.get_word Misc.PPC
            let sync_macro =Globals.get_sync_macro ()
          end in
	  let module PPC = PPCArch.Make(V) in
	  let module PPCLexParse = struct
	    type instruction = PPC.pseudo
	    type token = PPCParser.token

	    let lexer = PPCLexer.token
	    let parser = PPCParser.main
	  end in
          let module X = Make
              (PPC) (Tar) (PPCLexParse) (PPCCompile.Make(V)(Conf)) in
          X.compile hint avoid_cycle fst cycles name in_chan out_chan splitted
      | Misc.X86 ->
	  let module X86 = X86Arch.Make(V) in
	  let module X86LexParse = struct
	    type instruction = X86.pseudo
	    type token = X86Parser.token

	    let lexer = X86Lexer.token
	    let parser = X86Parser.main
	  end in
          let module X = Make
              (X86) (Tar) (X86LexParse) (X86Compile.Make(V)) in
          X.compile hint avoid_cycle fst cycles name in_chan out_chan splitted
      | Misc.ARM ->
	  let module ARM = ARMArch.Make(V) in
	  let module ARMLexParse = struct
	    type instruction = ARM.pseudo
	    type token = ARMParser.token

	    let lexer = ARMLexer.token
	    let parser = ARMParser.main
	  end in
          let module X = Make
              (ARM) (Tar) (ARMLexParse) (ARMCompile.Make(V)) in
          X.compile hint avoid_cycle fst cycles name in_chan out_chan splitted



    let from_file hint avoid_cycles fst cycles name out_chan =
      Misc.input_protect
        (fun in_chan ->
          from_chan hint avoid_cycles fst cycles name in_chan out_chan) name

    open MySys

    let report_failure name msg chan =
      if not (Globals.is_out ()) then begin
        let title = sprintf "!!! Test %s failed !!!" name in
        let warn = String.make (String.length title) '!' in
        output_line chan  warn ;
        output_line chan  title ;
        output_line chan  warn ;
        fprintf chan "Reported: %s\n" msg ;
        output_line chan ""
      end

    let read_no fname =
      Misc.input_protect
        (fun chan -> MySys.read_list chan (fun s -> Some s))
        fname

    let do_dont =
      let open Globals in
      match get_speedcheck () with
      | AllSpeed -> true
      | NoSpeed|SomeSpeed -> false


    let from_files names =  
      let avoid_cycles =
        let xs = match Globals.get_no () with
        | None -> []
        | Some fname -> read_no fname in
        let set = StringSet.of_list xs in
        fun cy -> StringSet.mem cy set in
      let hint = match Globals.get_hint () with
      | None -> Hint.empty
      | Some fname -> Hint.read fname in
      let out_chan =
        if Globals.is_out() then
          open_out (Tar.outname (MyName.outname "run" ".sh"))
        else stdout in
      let exp = match Globals.get_exp () with
      | None -> None
      | Some exp -> Some (open_out exp) in
      if Globals.is_out () then begin
        output_line out_chan "date" ;
        output_line out_chan "LITMUSOPTS=\"${@:-$LITMUSOPTS}\"" ;
        begin match Globals.get_para () with
        | Globals.Shell ->
            output_line out_chan "TMPDIR=${TMPDIR:-/var/tmp}" ;
            output_line out_chan "TMPDIR=$TMPDIR/tmp.$$" ;
            output_line out_chan "mkdir $TMPDIR"
        | Globals.Self -> ()
        end ;
        begin match Globals.get_crossrun () with
        | Globals.RemoteNone -> ()
        | Globals.RemoteAdb  ->
            fprintf out_chan "RDIR=%s\n" "/data/tmp" ;
            fprintf out_chan "adb shell mkdir $RDIR >/dev/null 2>&1\n" ;
            fprintf out_chan "dorun () {\n" ;
            fprintf out_chan "  EXE=$1\n" ;
            fprintf out_chan "  shift\n" ;
            fprintf out_chan "  OPTS=\"$@\"\n" ;
            fprintf out_chan "  BASE=$(basename $EXE)\n" ;
            fprintf out_chan "  adb push $EXE $RDIR >/dev/null 2>&1\n" ;
            fprintf out_chan "  adb shell $RDIR/$BASE $OPTS | tr -d '\\r' \n" ;
            fprintf out_chan "  adb shell rm $RDIR/$BASE >/dev/null 2>&1\n" ;
            fprintf out_chan "}\n" ;
        | Globals.RemoteHost h ->
            fprintf out_chan "RHOST=%s\n" h.Globals.host ;
            fprintf out_chan "RPORT=%i\n"
              (match h.Globals.port with| None -> 22| Some p -> p) ;
            fprintf out_chan "doscp () {\n" ;
            fprintf out_chan "  EXE=$1\n" ;
            fprintf out_chan "  scp -q -P $RPORT $EXE $RHOST: 2> /dev/null\n" ;
            fprintf out_chan "  while [ $? -ne 0 ]\n" ;
            fprintf out_chan "  do\n" ;    
            fprintf out_chan "    sleep 30\n" ;
            fprintf out_chan "    scp -q -P $RPORT $EXE $RHOST: 2> /dev/null\n" ;
            fprintf out_chan "  done\n" ;    
            fprintf out_chan "  true\n" ;
            fprintf out_chan "}\n" ;
            ()
        end ;
        let sleep = Globals.get_sleep () in
        if sleep >= 0 then fprintf out_chan "SLEEP=%i\n" sleep ;
        if do_dont then fprintf out_chan "LOG=./$$.txt\n"
      end ;
      let arch,sources,utils,_ =
        Misc.fold_argv
          (fun name (_,srcs,fst,cycles) ->
            match from_file hint avoid_cycles fst cycles name out_chan with
            | Completed (a,src,fst,cycles) ->
                begin match exp with
                | None -> ()
                | Some exp -> fprintf exp "%s\n" name
                end ;
                a,(src::srcs),fst,cycles
            | Absent a -> a,srcs,fst,cycles
            | Interrupted (a,e) ->
                let msg =  match e with
                | Misc.Exit -> "None"
                | Misc.Fatal msg 
                | Misc.UserError msg ->
                    Pos.pp_pos0 stderr name ;
                    Printf.eprintf "%s\n%!" msg ;
                    msg
                | e -> raise e  in
                report_failure name msg out_chan ;
                a,srcs,fst,cycles)
          names (Misc.X86,[],[],StringSet.empty) in

      begin match exp with
      | None -> ()
      | Some exp -> close_out exp
      end ;

      if (Globals.is_out ()) then begin
        match Globals.get_para () with
        | Globals.Shell -> output_line out_chan "rmdir $TMPDIR"
        | Globals.Self -> ()
      end ;
      if do_dont then output_line out_chan "rm -f $LOG" ;
      Run.report_machine arch out_chan ;
      if (Globals.is_out ()) then begin
        output_line out_chan "head -1 comp.sh" ;
        output_line out_chan "echo \"LITMUSOPTS=$LITMUSOPTS\"" ;
        output_line out_chan "date"
      end ;
      if (not (Globals.is_out ())) then begin
        List.iter MySys.remove utils        
      end ;
      flush out_chan ;


      if Globals.is_out () then begin
        close_out out_chan ;
        let sources = List.map Filename.basename  sources in
(* Shell script for sequential compilation *)
        Misc.output_protect
          (fun chan ->
            let gcc_opts = Run.get_gcc_opts arch in
            fprintf chan "GCC=%s\n" (Globals.get_gcc ()) ;
            fprintf chan "GCCOPTS=\"%s\"\n" gcc_opts ;
            let link_opts = Run.get_link_opts arch in
            fprintf chan "LINKOPTS=\"%s\"\n" link_opts ;
            fprintf chan "/bin/rm -f *.exe *.s\n" ;
            List.iter
              (fun s ->
                if Filename.check_suffix s ".c" then
                  fprintf chan "$GCC $GCCOPTS -O2 -c %s\n"
                    (Filename.basename s))
              utils ;
            let utils_objs =
              List.fold_right
                (fun s k ->
                  if Filename.check_suffix s ".c" then
                    let b = Filename.chop_suffix (Filename.basename s) ".c" in
                    (b ^ ".o") :: k
                  else k) utils [] in
            let utils_objs = String.concat " " utils_objs in
            List.iter
              (fun src ->
                let exe = Filename.chop_extension src ^ ".exe" in
(* No more moderate parallelism [blocks on abducens, sh bug ?] *)
                fprintf chan
                  "$GCC $GCCOPTS $LINKOPTS -o %s %s %s\n" exe utils_objs src ;
                let srcS = Filename.chop_extension src ^ ".s" in
                let srcT = Filename.chop_extension src ^ ".t" in
                fprintf chan "$GCC $GCCOPTS -S %s && awk -f show.awk %s > %s && /bin/rm %s\n"
                  src srcS srcT srcS ;
                ())
              (List.rev sources))
          (Tar.outname (MyName.outname "comp" ".sh")) ;
(* Add a small README file *)
        Misc.output_protect
          (fun chan ->
            GenerateCrossDoc.gen_readme_src chan arch sources)
          (Tar.outname (MyName.outname "README" ".txt")) ;
(* Makefile for parallel compilation *)
        Misc.output_protect
          (fun chan ->
(* Variables *)
            let gcc_opts = Run.get_gcc_opts arch in
            fprintf chan "GCC=%s\n" (Globals.get_gcc ()) ;
            fprintf chan "GCCOPTS=%s\n" gcc_opts ;
            let link_opts = Run.get_link_opts arch in
            fprintf chan "LINKOPTS=%s\n" link_opts ;
            fprintf chan "SRC=\\\n" ;
            List.iter
              (fun src -> fprintf chan " %s\\\n" src)
              (List.rev sources) ;
            fprintf chan "\n" ;
            fprintf chan "EXE=$(SRC:.c=.exe)\n" ;
            fprintf chan "T=$(SRC:.c=.t)\n" ;
            fprintf chan "\n" ;
(* Entry points *)
            fprintf chan "all: $(EXE) $(T)\n" ;
            fprintf chan "\n" ;
            fprintf chan "clean:\n" ;
            fprintf chan "\t/bin/rm -f *.o *.s *.t *.exe *~\n" ;
            fprintf chan "\n" ;
            fprintf chan "cleansource:\n" ;
            fprintf chan "\t/bin/rm -f *.o *.c *.h *.s *~\n" ;
            fprintf chan "\n" ;
(* Utilities *)
            let utils =
              List.fold_right
                (fun s k ->
                  if Filename.check_suffix s ".c" then
                    let b = Filename.chop_suffix (Filename.basename s) ".c" in
                    b :: k
                  else k) utils [] in          
            List.iter
              (fun u ->
                let src = u ^ ".c" and obj = u ^ ".o" in
                fprintf chan "%s: %s\n" obj src ;
                fprintf chan "\t$(GCC) $(GCCOPTS) -O2 -c %s\n" src ;
                fprintf chan "\n")
              utils ;
(* .exe pattern rule *)
            let utils_objs =
              String.concat " " (List.map (fun s -> s ^ ".o") utils) in
(* Cannot use %.s here for Mac that preprocesses assembly files,
   and because #comments makes the preprocessor crash ! *)
            let src_ext = match Globals.get_target_os arch with
            | Globals.Mac -> 'c'
            | Globals.Linux|Globals.AIX -> 's' in
            fprintf chan "%%.exe:%%.%c %s\n" src_ext utils_objs ;
            fprintf chan
              "\t$(GCC) $(GCCOPTS) $(LINKOPTS) -o $@ %s $<\n" utils_objs ;
            fprintf chan "\n" ;
(* .s pattern rule *)
            fprintf chan "%%.s:%%.c\n" ;
            fprintf chan "\t$(GCC) $(GCCOPTS) -S $<\n" ;
            fprintf chan "\n" ;
(* .t pattern rule *)
            fprintf chan "%%.t:%%.s\n" ;
            fprintf chan "\tawk -f show.awk $< > $@\n" ;
            fprintf chan "\n")
          (Tar.outname (MyName.outname "Makefile" "")) ;
        Tar.tar  () ;
        if Globals.cross_mode () then begin
          (* Normal tar *)
          (* Now build new directory structure *)
          let top = Filename.temp_file "dir" ".tmp" in
          MySys.mkdir top ;
          let dname = "litmus_tests" in
          let dir = Filename.concat top dname in
          MySys.mkdir dir ;
          (* Put this in dir *)
          Misc.output_protect 
            (fun chan -> output_line chan (GenerateCrossDoc.gen_makefile arch))
            (Filename.concat dir "Makefile") ;
          Misc.output_protect 
            (fun chan -> GenerateCrossDoc.gen_readme chan arch)
            (Filename.concat dir "README.txt") ;
(* Untar sources in src sub-directory *)
          if Tar.is_archive then begin
            let src = Filename.concat dir "src" in
            MySys.mkdir src ;
            let cmd =
              let tf = Filename.basename (Globals.get_tar ()) in
              sprintf "mv %s %s && cd %s && tar x%sf %s && /bin/rm -f %s"
                (Globals.get_tar ()) src src (Tar.tarz ()) tf tf in
            MySys.exec_stdout cmd ;
            Tar.tar_dir dir
          end else begin
            let tgt = Globals.get_tar () in
            let tmp = sprintf "%s.tmp" tgt in
            let src = Filename.concat tgt "src" in
            let com =
              sprintf
                "mv %s %s && mv %s %s && mv %s %s && /bin/rm -rf %s"
                tgt tmp dir tgt tmp src top in
            MySys.exec_stdout com
          end
        end
      end
  end    

